## Copyright (C) 2006-2012, Thomas Treichl <treichl@users.sourceforge.net>
## Copyright (C) 2013, Roberto Porcu' <roberto.porcu@polimi.it>
## OdePkg - A package for solving ordinary differential equations and more
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; If not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn  {Function File} {[@var{t}, @var{y}] =} odeRATTLE (@var{fun}, @var{constr}, @var{trange}, @var{init})
## @deftypefnx {Function File} {[@var{t}, @var{y}] =} odeRATTLE (@var{fun}, @var{constr}, @var{trange}, @var{init}, @var{ode_opt})
## @deftypefnx {Function File} {[@var{t}, @var{y}] =} odeRATTLE (@dots{}, @var{par1}, @var{par2}, @dots{})
## @deftypefnx {Function File} {[@var{t}, @var{y}, @var{te}, @var{ye}, @var{ie}] =} odeRATTLE (@dots{})
## @deftypefnx {Function File} {@var{solution} =} odeRATTLE (@dots{})
##
## This function is a geometric integrator for constrained Hamilton's equations
## of motion that uses the RATTLE method of order 2 to find the new solution,
## step by step.
##
## @var{fun} is a function_handle or an inline function that
## defines the system
## @ifhtml
## @example
## @math{q' =  dH/dp (t,[q;p])}
## @math{p' = -dH/dq (t,[q;p])}
## @end example
## @end ifhtml
## @ifnothtml
## @math{q' =  dH/dp (t,[q;p])}
## @math{p' = -dH/dq (t,[q;p])},
## @end ifnothtml
## where @math{H(t,[q;p])} is the Hamiltonian of the problem, @math{q} are
## generalized coordinates and @math{p} are generalized momenta. This function
## must take two input arguments: the time and the unknowns. It must return a
## vector or a cell (if the Hamiltonian Hessian is passed to speedup
## computations) containing as first element the Hamilton's equations of motion
## (the previous system of equations) and as second element the Hessian of the
## Hamiltonian.
##
## @var{constr} is a function_handle or an inline function
## that defines the constraints of the problem @math{g(q)=0}. This
## function must take one input argument and must return a vector or a cell (if
## the constraint gradient and constraint Hessian are passed) containing as
## first element the equations defining the holonomic constraints and as second
## and third elements respectively the gradient and the Hessian of this
## constraints function (this is not mandatory but it can speedup the
## computation and improve accuracy).
##
## @var{trange} specifies the time interval over which the ODE will be
## evaluated.  Typically, it is a two-element vector specifying the initial and
## final times (@code{[tinit, tfinal]}).  If there are more than two elements
## then the solution will also be evaluated at these intermediate time
## instances unless the integrate function specified is
## @command{integrate_n_steps}.
##
## By default, @code{odeRATTLE} uses an adaptive timestep with the
## @code{integrate_adaptive} algorithm.  The tolerance for the timestep
## computation may be changed by using the option @qcode{"Tau"}, that has a
## default value of @math{1e-6}.  If the ODE option @qcode{"TimeStepSize"} is
## not empty, then the stepper called will be @code{integrate_const}.  If, in
## addition, the option @qcode{"TimeStepNumber"} is also specified then the
## integrate function @code{integrate_n_steps} will be used.
##
## @var{init} contains the initial value for the unknowns. If it is a row or
## column vector containing the generalized coordinates in its first half and
## the generalized momenta in the second half. Each momentum correspond to the
## coordinate with the same relative position inside the vector of initial
## conditions.
##
## The optional fourth argument @var{ode_opt} specifies non-default options to
## the ODE solver. It is a structure generated by @code{odeset}.The option
## @var{"InexactSolver"} can be set to @var{'inexact_newton'} or not. If is set
## to @var{'inexact_newton'} in the stepper it will be used this inexact solver
## to solve the system. If is not set to @var{'inexact_newton'} then it will be
## used the solver @command{fsolve}. The @var{"UseJacobian"} option can be set
## on @var{'yes'} or @var{'no'}. In the first case, in the stepper, it will be
## used the Hamiltonian Hessian, the constraint gradient and the constraint
## Hessian if given by the user otherwise they will be approximated by the
## utility functions @command{approx_Hamilt_Hess}, @command{approx_Constr_grad}
## and @command{approx_Constr_Hess}. The @command{inexact_newton} solver has
## some options that can be set by set them in this fourth argument, they will
## be explained in the relative section.
##
## The function typically returns two outputs.  Variable @var{t} is a
## column vector and contains the times where the solution was found.  The
## output @var{y} is a matrix in which each column refers to a different
## unknown of the problem and each row corresponds to a time in @var{t}.
##
## The output can also be returned as a structure @var{solution} which
## has field @var{x} containing the time where the solution was evaluated and
## field @var{y} containing the solution matrix for the times in @var{x}.
## Use @code{fieldnames (@var{solution})} to see the other fields and additional
## information returned.
##
## @seealso{odeset,odeget,odepkg}
##
## @end deftypefn

function varargout = odeRATTLE (fun, constraint, trange, init, varargin)

  if (nargin < 4)
    print_usage ();
  endif

  order = 2;
  solver = 'odeRATTLE';

  if (nargin >= 5)
    if (! isstruct (varargin{1}))
      ## varargin{1:len} are parameters for fun
      odeopts = odeset ();
      odeopts.funarguments = varargin;
    elseif (length (varargin) > 1)
      ## varargin{1} is an OdePkg options structure opt
      odeopts = ode_struct_value_check ("odeRATTLE", varargin{1}, "odeRATTLE");
      odeopts.funarguments = {varargin{2:length(varargin)}};
    else ## if (isstruct (varargin{1}))
      odeopts = ode_struct_value_check ("odeRATTLE", varargin{1}, "odeRATTLE");
      odeopts.funarguments = {};
    endif
  else  # if (nargin == 3)
    odeopts = odeset ();
    odeopts.funarguments = {};
  endif

  if (! isvector (trange) || ! isnumeric (trange))
    error ("OdePkg:InvalidArgument",
           "odeRATTLE: TRANGE must be a numeric vector");
  endif

  TimeStepNumber = odeget (odeopts, "TimeStepNumber", [], "fast");
  TimeStepSize = odeget (odeopts, "TimeStepSize", [], "fast");
  if (length (trange) < 2
      && (isempty (TimeStepSize) || isempty (TimeStepNumber)))
    error ("OdePkg:InvalidArgument",
           "odeRATTLE: TRANGE must contain at least 2 elements");
  elseif (trange(1) == trange(2))
    error ("OdePkg:InvalidArgument",
           "odeRATTLE: invalid time span, TRANGE(1) == TRANGE(2)");
  else
    odeopts.direction = sign (trange(2) - trange(1));
  endif
  trange = trange(:);

  if (! isvector (init) || ! isnumeric (init))
    error ("OdePkg:InvalidArgument",
           "odeRATTLE: INIT must be a numeric vector");
  endif
  init = init(:);

  init_size = length (init);
  if (mod (init_size, 2) != 0)
    error("OdePkg:InvalidArgument",
          "odeRATTLE: INIT must be a vector with an even number of components");
  endif

  if (ischar (fun))
    try
      fun = str2func (fun);
    catch
      warning (lasterr);
    end_try_catch
  endif
  if (! isa (fun, "function_handle"))
    error ("OdePkg:InvalidArgument",
           "odeRATTLE: FUN must be a valid function handle");
  endif

  flag = 0;
  try
    fun (trange(1), init){1};
  catch
    flag = 1;
  end_try_catch

  if (flag == 0)
    fun_new = @(t,x) fun (t, x){1};
  else
    fun_new = fun;
  endif
  
  if ! (isa (constraint, "function_handle") || isa (constraint, "inline"))
    error ("OdePkg:InvalidArgument'",
           "odeRATTLE: FUN must be a valid function handle");
  endif

  warning ("OdePkg:InvalidArgument",
           ["odeRATTLE: solver assuming initial conditions vector contains", ...
           " coordinates in first half and momenta in second half"]);

  ## Start preprocessing, have a look which options are set in
  ## odeopts, check if an invalid or unused option is set
  if (isempty (TimeStepNumber) && isempty (TimeStepSize))
    integrate_func = "adaptive";
    odeopts.stepsizefixed = false;
  elseif (! isempty (TimeStepNumber) && ! isempty (TimeStepSize))
    integrate_func = "n_steps";
    odeopts.stepsizefixed = true;
    if (sign (TimeStepSize) != odeopts.direction)
      warning ("OdePkg:InvalidArgument",
               ["odeRATTLE: option 'TimeStepSize' has the wrong sign, ", ...
                "but will be corrected automatically\n"]);
      TimeStepSize = -TimeStepSize;
    endif
  elseif (isempty (TimeStepNumber) && ! isempty (TimeStepSize))
    integrate_func = "const";
    odeopts.stepsizefixed = true;
    if (sign (TimeStepSize) != odeopts.direction)
      warning ("OdePkg:InvalidArgument",
               ["odeRATTLE: option 'TimeStepSize' has the wrong sign, ",
                "but will be corrected automatically\n"]);
      TimeStepSize = -TimeStepSize;
    endif
  else
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: assuming an adaptive integrate function\n");
    integrate_func = "adaptive";
  endif

  if (isempty (odeopts.RelTol) && ! odeopts.stepsizefixed)
    odeopts.RelTol = 1e-3;
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'RelTol' not set, new value %f will be used\n",
             odeopts.RelTol);
  elseif (! isempty (odeopts.RelTol) && odeopts.stepsizefixed)
    warning ("OdePkg:InvalidArgument",
             ["odeRATTLE: option 'RelTol' is ignored", ...
              " when fixed time stamps are given\n"]);
  endif

  if (isempty (odeopts.AbsTol) && ! odeopts.stepsizefixed)
    odeopts.AbsTol = 1e-6;
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'AbsTol' not set, new value %f will be used\n",
             odeopts.AbsTol);
  elseif (! isempty (odeopts.AbsTol) && odeopts.stepsizefixed)
    warning ("OdePkg:InvalidArgument",
             ["odeRATTLE: option 'AbsTol' is ignored", ...
              " when fixed time stamps are given\n"]);
  else
    odeopts.AbsTol = odeopts.AbsTol(:);  # Create column vector
  endif

  odeopts.normcontrol = strcmp (odeopts.NormControl, "on");

  if (! isempty (odeopts.NonNegative))
    if (isempty (odeopts.Mass))
      odeopts.havenonnegative = true;
    else
      odeopts.havenonnegative = false;
      warning ("OdePkg:InvalidArgument",
               ["odeRATTLE: option 'NonNegative' is ignored", ...
                " when mass matrix is set\n"]);
    endif
  else
    odeopts.havenonnegative = false;
  endif

  if (isempty (odeopts.OutputFcn) && nargout == 0)
    odeopts.OutputFcn = @odeplot;
    odeopts.haveoutputfunction = true;
  else
    odeopts.haveoutputfunction = ! isempty (odeopts.OutputFcn);
  endif

  odeopts.haveoutputselection = ! isempty (odeopts.OutputSel);

  if (odeopts.Refine > 0)
    odeopts.haverefine = true;
  else
    odeopts.haverefine = false;
  endif

  if (isempty (odeopts.InitialStep) && strcmp (integrate_func, "adaptive"))
    odeopts.InitialStep = ...
      odeopts.direction * starting_stepsize (order, fun, trange(1),
                                                  init,
                                                  odeopts.AbsTol,
                                                  odeopts.RelTol,
                                                  odeopts.normcontrol);
    warning ("OdePkg:InvalidArgument",
             ["odeRATTLE: option 'InitialStep' not set,", ...
              " estimated value %f will be used\n"],
             odeopts.InitialStep);
  elseif (isempty (odeopts.InitialStep))
    odeopts.InitialStep = TimeStepSize;
  endif

  if (isempty (odeopts.MaxStep) && ! odeopts.stepsizefixed)
    odeopts.MaxStep = abs (trange(end) - trange(1)) / 10;
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'MaxStep' not set, new value %f will be used\n",
             odeopts.MaxStep);
  endif

  odeopts.haveeventfunction = ! isempty (odeopts.Events);

  ## The options "Jacobian", "JPattern" and "Vectorized" will be ignored
  ## by this solver because this solver uses an explicit Runge-Kutta method
  ## and therefore no Jacobian calculation is necessary.
  if (! isempty (odeopts.Jacobian))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'Jacobian' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.JPattern))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'JPattern' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.Vectorized))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'Vectorized' is ignored by this solver\n");
  endif

  ## Special options for RATTLE
  if (! isfield (odeopts, "InexactSolver") || isempty (odeopts.InexactSolver))
    odeopts.InexactSolver = "fsolve";
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'InexactSolver' not set, default value '%s' is used",
             odeopts.InexactSolver);
  endif

  if (! isfield (odeopts, "MaxNewtonIterations") || isempty (odeopts.MaxNewtonIterations))
    odeopts.MaxNewtonIterations = 100;
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'MaxNewtonIterations' not set, default value %i is used",
             odeopts.MaxNewtonIterations);
  endif

  if (! isfield (odeopts, "NewtonTol") || isempty (odeopts.NewtonTol))
    odeopts.NewtonTol = 1e-7;
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'NewtonTol' not set, default value %g is used",
             odeopts.NewtonTol);
  endif

  if (! isfield (odeopts, "UseJacobian") || isempty (odeopts.UseJacobian))
    odeopts.UseJacobian = "no";
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'UseJacobian' not set, default value '%s' is used",
             odeopts.UseJacobian);
  endif

  ## Other options that are not used by this solver.
  if (! isempty (odeopts.Mass))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'Mass' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.MStateDependence))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'MStateDependence' is ignored by this solver\n");
  end

  if (! isempty (odeopts.MvPattern))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'MvPattern' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.MassSingular))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'MassSingular' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.InitialSlope))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'InitialSlope' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.MaxOrder))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'MaxOrder' is ignored by this solver\n");
  endif

  if (! isempty (odeopts.BDF))
    warning ("OdePkg:InvalidArgument",
             "odeRATTLE: option 'BDF' is ignored by this solver\n");
  endif

  ## Starting the initialisation of the core solver odeRATTLE
  SubOpts = odeopts;

  dim = length (init) / 2;

  flag = 0;
  try
    constraint (init(1:dim)){1};
  catch
    flag = 1;
  end_try_catch

  if (flag == 0)
    constraint_new = @(q) constraint (q){1};
  else
    constraint_new = constraint;
  endif

  SubOpts.ConstraintFcn = constraint_new;
  SubOpts.ConstraintsNb = length (constraint_new (init(1:dim)));

  flag = 0;
  try
    constraint (init(1:dim)){2};
  catch
    flag = 1;
    warning ("OdePkg:InvalidArgument", ...
             "gradient of constraint will be approximated");
  end_try_catch

  if (flag == 1)
    GradConstrainFcn = @(q) approx_Constr_grad (constraint_new, q);
  else
    GradConstrainFcn = @(q) constraint (q){2};
  endif

  SubOpts.ConstraintGradFcn = GradConstrainFcn;

  use_Jacobian = odeget (odeopts, "UseJacobian", "no", "fast_not_empty");
  use_Jacobian = lower (deblank (use_Jacobian));
  if (strcmp (use_Jacobian, "yes"))
    flag = 0;
    try
      fun (trange(1), init){2};
    catch
      flag = 1;
      warning ("OdePkg:InvalidArgument", ...
               "consider returning the Hessian of Hamiltonian in function ", ...
               "''%s'', this may speedup the program", func2str (fun));
    end_try_catch

    if (flag == 1)
      Hessian = @(t,x) approx_Hamilt_Hess (fun_new, t, x);
    else
      Hessian = @(t,x) fun(t, x){2};
    endif

    SubOpts.HamiltonianHessFcn = Hessian;

    flag = 0;
    try
      constraint (init(1:dim)){3};
    catch
      flag = 1;
      warning ("OdePkg:InvalidArgument", ...
               "consider returning the Hessian of constraint in function ", ...
               "''%s'', this may speedup the program", func2str (constraint));
    end_try_catch

    if (flag == 1)
      ConstraintHess = @(q) approx_Constr_Hess (GradConstrainFcn, q);
    else
      ConstraintHess = @(q) constraint (q){3};
    endif

    SubOpts.ConstraintHessFcn = ConstraintHess;
  endif

  inexact_solver = odeget (odeopts, "InexactSolver", "fsolve", "fast_not_empty");
  inexact_solver = lower (deblank (inexact_solver));
  if (strcmp(inexact_solver, "inexact_newton"))
    SubOpts.solver = @inexact_newton;
  else
    SubOpts.solver = @fsolve;
  end
  
  switch integrate_func
    case "adaptive"
      SobOpts.Tau = odeget (odeopts, "Tau", 1.e-6, "fast_not_empty");
      solution = integrate_adaptive (@rattle, order, fun_new, trange,
                                     init, SubOpts);
    case "n_steps"
      solution = integrate_n_steps (@rattle, fun_new, trange(1), init,
                                    odeopts.TimeStepSize,
                                    odeopts.TimeStepNumber, SubOpts);
    case "const"
      solution = integrate_const (@rattle, fun_new, trange, init,
                                  odeopts.TimeStepSize, SubOpts);
  endswitch

  ## Postprocessing, do whatever when terminating integration algorithm
  if (odeopts.haveoutputfunction)  # Cleanup plotter
    feval (odeopts.OutputFcn, solution.t(end),
           solution.x(end,:)', "done", odeopts.funarguments{:});
  endif
  if (odeopts.haveeventfunction)   # Cleanup event function handling
    ode_event_handler (odeopts.Events, solution.t(end),
                         solution.x(end,:)', "done",
                         odeopts.funarguments{:});
  endif

  ## Print additional information if option Stats is set
  if (strcmp (odeopts.Stats, "on"))
    havestats = true;
    nsteps    = solution.vcntloop-2;              # vcntloop from 2..end
    nfailed   = (solution.vcntcycles-1)-nsteps+1; # vcntcycl from 1..end
    ndecomps  = 0;  # number of LU decompositions
    npds      = 0;  # number of partial derivatives
    nlinsols  = 0;  # no. of solutions of linear systems
    ## Print cost statistics if no output argument is given
    if (nargout == 0)
      printf ("Number of successful steps: %d\n", nsteps);
      printf ("Number of failed attempts:  %d\n", nfailed);
    endif
  else
    havestats = false;
  endif

  if (nargout == 2)
    varargout{1} = solution.t;      # Time stamps are first output argument
    varargout{2} = solution.x;      # Results are second output argument
  elseif (nargout == 1)
    varargout{1}.x = solution.t;    # Time stamps are saved in field x
    varargout{1}.y = solution.x;    # Results are saved in field y
    varargout{1}.solver = solver;   # Solver name is saved in field solver
    if (odeopts.haveeventfunction)
      varargout{1}.ie = solution.event{2};  # Index info which event occurred
      varargout{1}.xe = solution.event{3};  # Time info when an event occurred
      varargout{1}.ye = solution.event{4};  # Results when an event occurred
    endif
    if (havestats)
      varargout{1}.stats = struct ();
      varargout{1}.stats.nsteps   = nsteps;
      varargout{1}.stats.nfailed  = nfailed;
      varargout{1}.stats.nfevals  = nfevals;
      varargout{1}.stats.npds     = npds;
      varargout{1}.stats.ndecomps = ndecomps;
      varargout{1}.stats.nlinsols = nlinsols;
    endif
  elseif (nargout == 5)
    varargout = cell (1,5);
    varargout{1} = solution.t;
    varargout{2} = solution.x;
    if (odeopts.haveeventfunction)
      varargout{3} = solution.event{3};  # Time info when an event occurred
      varargout{4} = solution.event{4};  # Results when an event occurred
      varargout{5} = solution.event{2};  # Index info which event occurred
    endif
  endif

endfunction

%! # We are using the "armonic oscillator" implementation for all tests that
%! # are done for this function.
%!
%!function [ydot] = pendulum (vt, vy)
%!  ydot = [0.5*vy(length(vy)/2+1:end); -(2*5*9.806)*[zeros(length(vy)/4);ones(length(vy)/4)]];
%!endfunction
%!function [ydot] = pendulum_with_H (vt, vy)
%!  Hamilton = [0.5*vy(length(vy)/2+1:end); -(2*5*9.806)*[zeros(length(vy)/4);ones(length(vy)/4)]];
%!  Hess = [zeros(length(vy)/2,length(vy));zeros(length(vy)/2),0.5*eye(length(vy)/2)];
%!  ydot = {Hamilton,Hess};
%!endfunction
%!function [g] = constraint (vy)
%!  constr = (vy(:)')*vy(:) - 25.0;
%!  grad = 2.0*vy(:)';
%!  g = {constr,grad};
%!endfunction
%!
%!shared m, L, g
%!  m = 2.0; 
%!  L = 5.0; 
%!  g = 9.806;
%!
%! ## Turn off output of warning messages for all tests, turn them on
%! ## again if the last test is called
%!error ## output argument
%!  warning ('off', 'OdePkg:InvalidArgument');
%!  B = odeRATTLE (1, @constraint, [0 1], [L*cos(-pi/4) L*sin(-pi/4) 0 0]);
%!error ## input argument number one
%!  [vt, vy] = odeRATTLE (1, @constraint, [0 1], [L*cos(-pi/4) L*sin(-pi/4) 0 0]);
%!error ## input argument number two
%!  [vt, vy] = odeRATTLE (@pendulum, 2, [0 1], [L*cos(-pi/4) L*sin(-pi/4) 0 0]);
%!error ## input argument number three
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint, 0, [L*cos(-pi/4) L*sin(-pi/4) 0 0]);
%!error ## input argument number four
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint,  [0 1], [L*cos(-pi/4) L*sin(-pi/4) 0 0 0]);
%!test ## two output arguments
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint, [0 0.4], [L*cos(-pi/4) L*sin(-pi/4) 0 0]);
%!  assert ([vt(end), sqrt(vy(end,1)^2+vy(end,2)^2)], [0.4, L], 1e-3);
%!test ## extra input arguments passed through
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint, [0 0.4], [L*cos(-pi/4) L*sin(-pi/4) 0 0], 12, 13, 'KL');
%!  assert ([vt(end), sqrt(vy(end,1)^2+vy(end,2)^2)], [0.4, L], 1e-2);
%!test ## empty OdePkg structure *but* extra input arguments
%!  opt = odeset();
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint, [0 0.4], [L*cos(-pi/4) L*sin(-pi/4) 0 0], opt, 12, 13, 'KL');
%!  assert ([vt(end), sqrt(vy(end,1)^2+vy(end,2)^2)], [0.4, L], 1e-2);
%!test  # Solve with intermidiate step
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint, [0 0.2 0.4], [L*cos(-pi/4) L*sin(-pi/4) 0 0]);
%!  assert ([vt(end), sqrt(vy(end,1)^2+vy(end,2)^2)], [0.4, L], 1e-2);
%!test ## use_jacobian option
%!  opt = odeset();
%!  opt.UseJacobian = 'yes';
%!  [vt, vy] = odeRATTLE (@pendulum_with_H, @constraint, [0 0.4], [L*cos(-pi/4) L*sin(-pi/4) 0 0], opt);
%!  assert ([vt(end), sqrt(vy(end,1)^2+vy(end,2)^2)], [0.4, L], 1e-2);
%!test ## use approximated jacobian
%!  opt = odeset();
%!  opt.UseJacobian = 'yes';
%!  [vt, vy] = odeRATTLE (@pendulum, @constraint, [0 0.4], [L*cos(-pi/4) L*sin(-pi/4) 0 0], opt);
%!  assert ([vt(end), sqrt(vy(end,1)^2+vy(end,2)^2)], [0.4, L], 1e-2);
%!
%!  warning ('on', 'OdePkg:InvalidArgument');

## Local Variables: ***
## mode: octave ***
## End: ***
