%# Copyright (C) 2006-2012, Thomas Treichl <treichl@users.sourceforge.net>
%# Copyright (C) 2013, Roberto Porcu' <roberto.porcu@polimi.it>
%# OdePkg - A package for solving ordinary differential equations and more
%#
%# This program is free software; you can redistribute it and/or modify
%# it under the terms of the GNU General Public License as published by
%# the Free Software Foundation; either version 2 of the License, or
%# (at your option) any later version.
%#
%# This program is distributed in the hope that it will be useful,
%# but WITHOUT ANY WARRANTY; without even the implied warranty of
%# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%# GNU General Public License for more details.
%#
%# You should have received a copy of the GNU General Public License
%# along with this program; If not, see <http://www.gnu.org/licenses/>.

%# -*- texinfo -*-
%# @deftypefn {Command} {[@var{sol}] =} odeSPVI (@var{@@fun}, @var{slot}, @var{init}, [@var{opt}], [@var{par1}, @var{par2}, @dots{}])
%# @deftypefnx {Command} {[@var{t}, @var{y}, [@var{xe}, @var{ye}, @var{ie}]] =} odeSPVI (@var{@@fun}, @var{slot}, @var{init}, [@var{opt}], [@var{par1}, @var{par2}, @dots{}])
%# 
%# This function is a geometric integrator for Hamilton's equations of motion that uses the spectral variational integrators method to find the new solution, step by step.
%# 
%# This function can be called with two output arguments: @var{t} and @var{y}. Variable @var{t} is a column vector and contains the time stamps, instead @var{y} is a matrix in which each column refers to a different unknown of the problem and the rows number is the same of @var{t} rows number so that each row of @var{y} contains the values of all unknowns at the time value contained in the corresponding row in @var{t}.
%# 
%# The first input argument must be a function_handle or an inline function that defines the system
%# @ifhtml
%# @example
%# @math{q' =  dH/dp (t,[q;p])}
%# @math{p' = -dH/dq (t,[q;p])}
%# @end example
%# @end ifhtml
%# @ifnothtml
%# @math{q' =  dH/dp (t,[q;p])}
%# @math{p' = -dH/dq (t,[q;p])},
%# @end ifnothtml
%# where @math{H(t,[q;p])} is the Hamiltonian of the problem, @math{q} are generalized coordinates and @math{p} are generalized momenta. As described above, this function must take two input arguments, where the first is the time and the second the unknowns, and must return a vector or (if you want to pass also the Hamiltonian Hessian) a cell containing as first element the Hamilton's equations of motion (the previous system of equations) and as second element the Hessian of the Hamiltonian (this is not mandatory but it can speedup the computation).
%# 
%# The second input argument must contain time informations. Usually it should be a vector with at least two elements which define the initial and the final time instants; if the elements are more than two, then the solution will be evaluated also at these intermediate time instants unless the integrate function called is the @command{integrate_n_steps}. If there is only one time value, then it will give an error unless the options structure has no empty fields named @var{"TimeStepNumber"} and @var{"TimeStepSize"}. If the option @var{"TimeStepSize"} is not empty, then the stepper called will be @command{integrate_const}, if also @var{"TimeStepNumber"} is not empty it will be called the integrate function @command{integrate_n_steps}, otherwise it will be called @command{integrate_adaptive}. For this last possibility the user can set the tolerance for the timestep computation by setting a value to the option @var{"Tau"}, that as default value has @math{1.e-6}.
%# 
%# The third input argument must contain the initial value for the unknown. This must contain the generalized coordinates in its first half and the generalized momenta in the second half. Each momentum correspond to the coordinate with the same relative position inside the vector of initial conditions.
%# 
%# The fourth input argument is not mandatory and it should contain a structure with valid ODE fields. The @var{"PolynomialDegree"} and @var{"QuadratureOrder"} options can be used to set the maximum degree of the polynomials of the basis function and the order of the Gauss quadrature rule. If not set they are assumed equal respectively to 1 and 2. The option @var{"InexactSolver"} can be set to @var{'inexact_newton'} or not. If is set to @var{'inexact_newton'} in the stepper it will be used this inexact solver to solve the system. If is not set to @var{'inexact_newton'} then it will be used the solver @command{fsolve}. The @var{"UseJacobian"} option can be set on @var{'yes'} or @var{'no'}. In the first case, in the stepper, it will be used the Hamiltonian Hessian given by the user, otherwise it will be approximated by the utility function @command{approx_Hamilt_Hess}. The @var{'inexact_newton'} solver has some options that can be set by set them in this fourth argument, they will be explained in the relative section.
%# @end deftypefn
%#
%# @seealso{odepkg}

function [varargout] = odeSPVI (vfun, vslot, vinit, varargin)

  vsolver = 'odeSPVI';

  if (nargin == 0) %# Check number and types of all input arguments
    help (vsolver);
    error ('OdePkg:InvalidArgument', ...
      'number of input arguments must be greater than zero');
  end

  if (nargin < 3)
    print_usage;
  end
  
  if (nargin >= 4)
    if (~isstruct (varargin{1}))
      %# varargin{1:len} are parameters for vfun
      vodeoptions = odeset;
      vodeoptions.vfunarguments = varargin;
    elseif (length (varargin) > 1)
      %# varargin{1} is an OdePkg options structure vopt
      vodeoptions = odepkg_structure_check (varargin{1}, 'odeSPVI');
      vodeoptions.vfunarguments = {varargin{2:length(varargin)}};
    else %# if (isstruct (varargin{1}))
      vodeoptions = odepkg_structure_check (varargin{1}, 'odeSPVI');
      vodeoptions.vfunarguments = {};
    end
  else %# if (nargin == 3)
    vodeoptions = odeset; 
    vodeoptions.vfunarguments = {};
  end

  if (~isvector (vslot) || ~isnumeric(vslot))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  end

  if(length(vslot)<2 && (isempty(vodeoptions.TimeStepSize) || isempty(vodeoptions.TimeStepNumber) ))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  elseif(vslot(2)==vslot(1))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  else
    vodeoptions.vdirection = sign(vslot(2)-vslot(1));
  end
  vslot = vslot(:);

  if (~isvector (vinit) || ~isnumeric (vinit))
    error ('OdePkg:InvalidArgument', ...
      'third input argument must be a valid numerical value');
  end
  vinit = vinit(:);

  vinit_size = length(vinit);
  if(mod(vinit_size,2) != 0)
    error ('OdePkg:InvalidArgument', ...
      'third input argument must be a vector with an even number of components');
  end

  if ~(isa (vfun, 'function_handle') || isa (vfun, 'inline'))
    error ('OdePkg:InvalidArgument', ...
      'first input argument must be a valid function handle');
  end

  flag = 0;
  try
    vfun(vslot(1),vinit){1};
  catch
    flag = 1;
  end

  if(flag==0)
    vfun_new = @(t,x)vfun(t,x){1};
  else
    vfun_new = vfun;
  end

  warning('OdePkg:InvalidArgument', ...
      'solver assuming initial conditions vector contains coordinates in first half and momenta in second half');

  if( ~isfield(vodeoptions,'PolynomialDegree') )
    vorder = 2;
    warning ('OdePkg:InvalidArgument', ...
      'option "PolynomialDegree" not set, default value %d is used', 1);
  else
    vorder = 2*odeget(vodeoptions,'PolynomialDegree',1,'fast_not_empty');
  end

  %# Start preprocessing, have a look which options are set in
  %# vodeoptions, check if an invalid or unused option is set
  if(isempty(vodeoptions.TimeStepNumber) && isempty(vodeoptions.TimeStepSize))
    integrate_func = 'adaptive';
    vodeoptions.vstepsizefixed = false;
  elseif(~isempty(vodeoptions.TimeStepNumber) && ~isempty(vodeoptions.TimeStepSize))
    integrate_func = 'n_steps';
    vodeoptions.vstepsizefixed = true;
    if(sign(vodeoptions.TimeStepSize)!=vodeoptions.vdirection)
      warning('OdePkg:InvalidArgument', ...
        'option "TimeStepSize" has a wrong sign, it will be corrected automatically');
      vodeoptions.TimeStepSize = (-1)*vodeoptions.TimeStepSize;
    end
  elseif(isempty(vodeoptions.TimeStepNumber) && ~isempty(vodeoptions.TimeStepSize))
    integrate_func = 'const';
    vodeoptions.vstepsizefixed = true;
    if(sign(vodeoptions.TimeStepSize)!=vodeoptions.vdirection)
      warning('OdePkg:InvalidArgument', ...
        'option "TimeStepSize" has a wrong sign, it will be corrected automatically');
      vodeoptions.TimeStepSize = (-1)*vodeoptions.TimeStepSize;
    end
  else
    warning ('OdePkg:InvalidArgument', ...
      'assuming an adaptive integrate function');
    integrate_func = 'adaptive';
  end

  %# Get the default options that can be set with 'odeset' temporarily
  vodetemp = odeset;

  %# Implementation of the option RelTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.RelTol) && ~vodeoptions.vstepsizefixed)
    vodeoptions.RelTol = 1e-3;
    warning ('OdePkg:InvalidArgument', ...
      'Option "RelTol" not set, new value %f is used', vodeoptions.RelTol);
  elseif (~isempty (vodeoptions.RelTol) && vodeoptions.vstepsizefixed)
    warning ('OdePkg:InvalidArgument', ...
      'Option "RelTol" will be ignored if fixed time stamps are given');
  end

  %# Implementation of the option AbsTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.AbsTol) && ~vodeoptions.vstepsizefixed)
    vodeoptions.AbsTol = 1e-6;
    warning ('OdePkg:InvalidArgument', ...
      'Option "AbsTol" not set, new value %f is used', vodeoptions.AbsTol);
  elseif (~isempty (vodeoptions.AbsTol) && vodeoptions.vstepsizefixed)
    warning ('OdePkg:InvalidArgument', ...
      'Option "AbsTol" will be ignored if fixed time stamps are given');
  else
    vodeoptions.AbsTol = vodeoptions.AbsTol(:); %# Create column vector
  end

  %# Implementation of the option NormControl has been finished. This
  %# option can be set by the user to another value than default value.
  if (strcmp (vodeoptions.NormControl, 'on')) 
    vodeoptions.vnormcontrol = true;
  else 
    vodeoptions.vnormcontrol = false; 
  end

  %# Implementation of the option NonNegative has been finished. This
  %# option can be set by the user to another value than default value.
  if (~isempty (vodeoptions.NonNegative))
    if (isempty (vodeoptions.Mass))
      vodeoptions.vhavenonnegative = true;
    else
      vodeoptions.vhavenonnegative = false;
      warning ('OdePkg:InvalidArgument', ...
        'Option "NonNegative" will be ignored if mass matrix is set');
    end
  else 
    vodeoptions.vhavenonnegative = false;
  end

  %# Implementation of the option OutputFcn has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (vodeoptions.OutputFcn) && nargout == 0)
    vodeoptions.OutputFcn = @odeplot;
    vodeoptions.vhaveoutputfunction = true;
  elseif (isempty (vodeoptions.OutputFcn))
    vodeoptions.vhaveoutputfunction = false;
  else 
    vodeoptions.vhaveoutputfunction = true;
  end

  %# Implementation of the option OutputSel has been finished. This
  %# option can be set by the user to another value than default value.
  if (~isempty (vodeoptions.OutputSel))
    vodeoptions.vhaveoutputselection = true;
  else 
    vodeoptions.vhaveoutputselection = false; 
  end

  %# Implementation of the option OutputSave has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (vodeoptions.OutputSave))
    vodeoptions.OutputSave = 1;
  end

  %# Implementation of the option Refine has been finished. This option
  %# can be set by the user to another value than default value.
  if (vodeoptions.Refine > 0)
    vodeoptions.vhaverefine = true;
  else 
    vodeoptions.vhaverefine = false;
  end

  %# Implementation of the option Stats has been finished. This option
  %# can be set by the user to another value than default value.

  %# Implementation of the option InitialStep has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (vodeoptions.InitialStep) && strcmp(integrate_func,'adaptive'))
    vodeoptions.InitialStep = vodeoptions.vdirection* ...
      starting_stepsize (vorder, vfun_new, vslot(1), vinit, vodeoptions.AbsTol, vodeoptions.RelTol, vodeoptions.vnormcontrol);
    warning ('OdePkg:InvalidArgument', ...
      'option "InitialStep" not set, estimated value %f is used', vodeoptions.InitialStep);
  elseif(isempty (vodeoptions.InitialStep))
    vodeoptions.InitialStep = odeget(vodeoptions,'TimeStepSize');
  end

  %# Implementation of the option MaxStep has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.MaxStep) && ~vodeoptions.vstepsizefixed)
    vodeoptions.MaxStep = abs (vslot(2) - vslot(1)) / 10;
    warning ('OdePkg:InvalidArgument', ...
      'Option "MaxStep" not set, new value %f is used', vodeoptions.MaxStep);
  end

  %# Implementation of the option Events has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isempty (vodeoptions.Events))
    vodeoptions.vhaveeventfunction = true;
  else 
    vodeoptions.vhaveeventfunction = false;
  end

  %# The options 'Jacobian', 'JPattern' and 'Vectorized' will be ignored
  %# by this solver because this solver uses an explicit Runge-Kutta
  %# method and therefore no Jacobian calculation is necessary
  if (~isequal (vodeoptions.Jacobian, vodetemp.Jacobian))
    warning ('OdePkg:InvalidArgument', ...
      'option "Jacobian" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.JPattern, vodetemp.JPattern))
    warning ('OdePkg:InvalidArgument', ...
      'option "JPattern" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.Vectorized, vodetemp.Vectorized))
    warning ('OdePkg:InvalidArgument', ...
      'option "Vectorized" will be ignored, not yet implemented functionality');
  end

  %# Implementation of the option MaxNewtonIterations has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.MaxNewtonIterations))
    vodeoptions.MaxNewtonIterations = 100;
    warning ('OdePkg:InvalidArgument', ...
      'option "MaxNewtonIterations" not set, default value %d is used', vodeoptions.MaxNewtonIterations);
  end

  %# Implementation of the option NewtonTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.NewtonTol))
    vodeoptions.NewtonTol = 1e-7;
    warning ('OdePkg:InvalidArgument', ...
      'option "NewtonTol" not set, default value %g is used', vodeoptions.NewtonTol);
  end

  %# Implementation of the option Mass has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isequal (vodeoptions.Mass, vodetemp.Mass))
    warning ('OdePkg:InvalidArgument', ...
      'option "Mass" will be ignored by this solver');
  end

  %# Implementation of the option MStateDependence has been finished.
  %# This option can be set by the user to another value than default
  %# value. 
  if (~isequal (vodeoptions.MStateDependence, vodetemp.MStateDependence))
    warning ('OdePkg:InvalidArgument', ...
      'option "MStateDependence" will be ignored by this solver');
  end

  %# Other options that are not used by this solver. Print a warning
  %# message to tell the user that the option(s) is/are ignored.
  if (~isequal (vodeoptions.MvPattern, vodetemp.MvPattern))
    warning ('OdePkg:InvalidArgument', ...
      'option "MvPattern" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.MassSingular, vodetemp.MassSingular))
    warning ('OdePkg:InvalidArgument', ...
      'option "MassSingular" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.InitialSlope, vodetemp.InitialSlope))
    warning ('OdePkg:InvalidArgument', ...
      'option "InitialSlope" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.MaxOrder, vodetemp.MaxOrder))
    warning ('OdePkg:InvalidArgument', ...
      'option "MaxOrder" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.BDF, vodetemp.BDF))
    warning ('OdePkg:InvalidArgument', ...
      'option "BDF" will be ignored by this solver');
  end

  %# Starting the initialisation of the core solver odeSPVI
  q_dofs = vorder/2 + 1;

  if( ~isfield(vodeoptions,'QuadratureOrder') )
    p_dofs = q_dofs;
    warning ('OdePkg:InvalidArgument', ...
      'option "QuadratureOrder" not set, default value %d is used', p_dofs);
  else
    p_dofs = odeget(vodeoptions,'QuadratureOrder',q_dofs,'fast_not_empty');
  end

  if( q_dofs > p_dofs )
    error('OdePkg:InvalidArgument', ...
      'for the correct computation of the solution polynomial degree must be lower than quadrature order');
  end

  [nodes,weights,Legendre,derivatives] = golub_welsch(q_dofs,p_dofs);
  extreme_values = [(-1).^[0:q_dofs-1];ones(1,q_dofs)];

  q_dofs_err = q_dofs+1;
  p_dofs_err = p_dofs+1;
  [nodes_err,weights_err,Legendre_err,derivatives_err] = golub_welsch(q_dofs_err,p_dofs_err);
  extreme_values_err = [(-1).^[0:q_dofs_err-1];ones(1,q_dofs_err)];

  SubOpts = vodeoptions;

  SubOpts.Q_DoFs = q_dofs;
  SubOpts.P_DoFs = p_dofs;
  SubOpts.Nodes = nodes;
  SubOpts.Weights = weights;
  SubOpts.Legendre = Legendre;
  SubOpts.Derivatives = derivatives;
  SubOpts.Extremes = extreme_values;

  SubOpts.Q_DoFs_err = q_dofs_err;
  SubOpts.P_DoFs_err = p_dofs_err;
  SubOpts.Nodes_err = nodes_err;
  SubOpts.Weights_err = weights_err;
  SubOpts.Legendre_err = Legendre_err;
  SubOpts.Derivatives_err = derivatives_err;
  SubOpts.Extremes_err = extreme_values_err;

  use_Jacobian = odeget(vodeoptions,'UseJacobian','no','fast_not_empty');
  use_Jacobian = lower(deblank(use_Jacobian));

  if( strcmp(use_Jacobian,'yes') )
    flag = 0;
    try
      vfun(vslot(1),vinit){2};
    catch
      flag = 1;
      warning('OdePkg:InvalidArgument', ...
        'consider returning the Hessian of Hamiltonian in function ''%s'', this may speedup the program',func2str(vfun));
    end

    if( flag==1 )
      Hessian = @(t,x)approx_Hamilt_Hess(vfun_new,t,x);
    else
      Hessian = @(t,x)vfun(t,x){2};
    end

    SubOpts.HamiltonianHessFcn = Hessian;
  end

  inexact_solver = odeget(vodeoptions,'InexactSolver','fsolve','fast_not_empty');
  inexact_solver = lower(deblank(inexact_solver));
  if( strcmp(inexact_solver,'inexact_newton') )
    SubOpts.solver = @inexact_newton;
  else
    SubOpts.solver = @fsolve;
  end
  
  switch integrate_func
    case 'adaptive'
      solution = integrate_adaptive(@spectral_var_int,vorder,vfun_new,vslot,vinit,SubOpts);
    case 'n_steps'
      solution = integrate_n_steps(@spectral_var_int,vfun_new,vslot(1),vinit,vodeoptions.TimeStepSize,vodeoptions.TimeStepNumber,SubOpts);
    case 'const'
      solution = integrate_const(@spectral_var_int,vfun_new,vslot,vinit,vodeoptions.TimeStepSize,SubOpts);
  end

  %# Postprocessing, do whatever when terminating integration algorithm
  if (vodeoptions.vhaveoutputfunction) %# Cleanup plotter
    feval (vodeoptions.OutputFcn, solution.t(end), ...
      solution.x(end,:)', 'done', vodeoptions.vfunarguments{:});
  end
  if (vodeoptions.vhaveeventfunction)  %# Cleanup event function handling
    odepkg_event_handle (vodeoptions.Events, solution.t(end), ...
      solution.x(end,:)', 'done', vodeoptions.vfunarguments{:});
  end

  %# Print additional information if option Stats is set
  if (strcmp (vodeoptions.Stats, 'on'))
    vhavestats = true;
    vnsteps    = solution.vcntloop-2;                    %# vcntloop from 2..end
    vnfailed   = (solution.vcntcycles-1)-(solution.vcntloop-2)+1; %# vcntcycl from 1..end
    vndecomps  = 0;                             %# number of LU decompositions
    vnpds      = 0;                             %# number of partial derivatives
    vnlinsols  = 0;                             %# no. of solutions of linear systems
    %# Print cost statistics if no output argument is given
    if (nargout == 0)
      vmsg = fprintf (1, 'Number of successful steps: %d\n', vnsteps);
      vmsg = fprintf (1, 'Number of failed attempts:  %d\n', vnfailed);
    end
  else
    vhavestats = false;
  end

  if (nargout == 1)                 %# Sort output variables, depends on nargout
    varargout{1}.x = solution.t;   %# Time stamps are saved in field x
    varargout{1}.y = solution.x; %# Results are saved in field y
    varargout{1}.solver = vsolver;  %# Solver name is saved in field solver
    if (vodeoptions.vhaveeventfunction) 
      varargout{1}.ie = solution.vevent{2};  %# Index info which event occured
      varargout{1}.xe = solution.vevent{3};  %# Time info when an event occured
      varargout{1}.ye = solution.vevent{4};  %# Results when an event occured
    end
    if (vhavestats)
      varargout{1}.stats = struct;
      varargout{1}.stats.nsteps   = vnsteps;
      varargout{1}.stats.nfailed  = vnfailed;
      varargout{1}.stats.npds     = vnpds;
      varargout{1}.stats.ndecomps = vndecomps;
      varargout{1}.stats.nlinsols = vnlinsols;
    end
  elseif (nargout == 2)
    varargout{1} = solution.t;     %# Time stamps are first output argument
    varargout{2} = solution.x;   %# Results are second output argument
  elseif (nargout == 5)
    varargout{1} = solution.t;     %# Same as (nargout == 2)
    varargout{2} = solution.x;   %# Same as (nargout == 2)
    varargout{3} = [];              %# LabMat doesn't accept lines like
    varargout{4} = [];              %# varargout{3} = varargout{4} = [];
    varargout{5} = [];
    if (vodeoptions.vhaveeventfunction) 
      varargout{3} = solution.vevent{3};     %# Time info when an event occured
      varargout{4} = solution.vevent{4};     %# Results when an event occured
      varargout{5} = solution.vevent{2};     %# Index info which event occured
    end
  end

end

%! # We are using the "armonic oscillator" implementation for all tests that
%! # are done for this function.
%!function [ydot] = armonic (vt, vy)
%! ydot = [vy(length(vy)/2+1:end); -vy(1:length(vy)/2)];
%!endfunction
%!function [ydot] = armonic_with_H (vt, vy)
%! Hamilton = [vy(length(vy)/2+1:end); -vy(1:length(vy)/2)];
%! Hess = eye(length(vy));
%! ydot = {Hamilton,Hess};
%!endfunction
%!
%! %# Turn off output of warning messages for all tests, turn them on
%! %# again if the last test is called
%!error %# output argument
%!  warning ('off', 'OdePkg:InvalidArgument');
%!  B = odeSPVI (1, [1 5], [1 0], odeset('PolynomialDegree',2,'QuadratureOrder',3));
%!error %# input argument number one
%!  [vt, vy] = odeSPVI (1, [1 5], [1 0], odeset('PolynomialDegree',2,'QuadratureOrder',3));
%!error %# input argument number two
%!  [vt, vy] = odeSPVI (@armonic, 1, [1 0], odeset('PolynomialDegree',2,'QuadratureOrder',3));
%!error %# input argument number three
%!  [vt, vy] = odeSPVI (@armonic, [1 5], [1 0 0], odeset('PolynomialDegree',2,'QuadratureOrder',3));
%!test %# two output arguments
%!  [vt, vy] = odeSPVI (@armonic, [0 1], [1 0] ,odeset('PolynomialDegree',2,'QuadratureOrder',3));
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# extra input arguments passed through
%!  [vt, vy] = odeSPVI (@armonic, [0 1], [1 0], 12, 13, 'KL');
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# empty OdePkg structure *but* extra input arguments
%!  vopt = odeset();
%!  [vt, vy] = odeSPVI (@armonic, [0 1], [1 0], vopt, 12, 13, 'KL');
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# use_jacobian option
%!  vopt = odeset('PolynomialDegree',2,'QuadratureOrder',3,'UseJacobian','yes');
%!  [vt, vy] = odeSPVI (@armonic_with_H, [0 1], [1 0], vopt);
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# use approximated jacobian
%!  vopt = odeset('PolynomialDegree',2,'QuadratureOrder',3,'UseJacobian','yes');
%!  [vt, vy] = odeSPVI (@armonic, [0 1], [1 0], vopt);
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# InitialStep option
%!  vopt = odeset ('InitialStep', 1e-8);
%!  [vt, vy] = odeSPVI (@armonic, [0 0.2], [1 0], vopt);
%!  assert ([vt(4)-vt(1)], [1e-8], 1e-9);
%!
%!  warning ('on', 'OdePkg:InvalidArgument');

%# Local Variables: ***
%# mode: octave ***
%# End: ***
