%# Copyright (C) 2006-2012, Thomas Treichl <treichl@users.sourceforge.net>
%# Copyright (C) 2013, Roberto Porcu' <roberto.porcu@polimi.it>
%# OdePkg - A package for solving ordinary differential equations and more
%#
%# This program is free software; you can redistribute it and/or modify
%# it under the terms of the GNU General Public License as published by
%# the Free Software Foundation; either version 2 of the License, or
%# (at your option) any later version.
%#
%# This program is distributed in the hope that it will be useful,
%# but WITHOUT ANY WARRANTY; without even the implied warranty of
%# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%# GNU General Public License for more details.
%#
%# You should have received a copy of the GNU General Public License
%# along with this program; If not, see <http://www.gnu.org/licenses/>.

%# -*- texinfo -*-
%# @deftypefn {Command} {[@var{sol}] =} odeVV (@var{@@fun}, @var{slot}, @var{init}, [@var{opt}], [@var{par1}, @var{par2}, @dots{}])
%# @deftypefnx {Command} {[@var{t}, @var{y}, [@var{xe}, @var{ye}, @var{ie}]] =} odeVV (@var{@@fun}, @var{slot}, @var{init}, [@var{opt}], [@var{par1}, @var{par2}, @dots{}])
%# 
%# This function is a geometric integrator for Newton's equations of motion that uses the Velocity-Verlet method to find the new solution, step by step.
%# 
%# This function can be called with two output arguments: @var{t} and @var{y}. Variable @var{t} is a column vector and contains the time stamps, instead @var{y} is a matrix in which each column refers to a different unknown of the problem and the rows number is the same of @var{t} rows number so that each row of @var{y} contains the values of all unknowns at the time value contained in the corresponding row in @var{t}.
%# 
%# The first input argument must be a function_handle or an inline function that defines the system
%# @ifhtml
%# @example
%# @math{q' =  v (t,q)}
%# @math{v' =  a (t,q)}
%# @end example
%# @end ifhtml
%# @ifnothtml
%# @math{q' =  v (t,q)}
%# @math{v' =  a (t,q)},
%# @end ifnothtml
%# where @math{a(t,q)} is the acceleration of the system, @math{q} are the generalized coordinates and @math{v} are the generalized velocities. As described above, this function must take two input arguments, where the first is the time and the second the unknowns, and must return a cell or a vector containing the equations of the previous system.
%# 
%# The second input argument must contain time informations. Usually it should be a vector with at least two elements which define the initial and the final time instants; if the elements are more than two, then the solution will be evaluated also at these intermediate time instants unless the integrate function called is the @command{integrate_n_steps}. If there is only one time value, then it will give an error unless the options structure has no empty fields named @var{"TimeStepNumber"} and @var{"TimeStepSize"}. If the option @var{"TimeStepSize"} is not empty, then the stepper called will be @command{integrate_const}, if also @var{"TimeStepNumber"} is not empty it will be called the integrate function @command{integrate_n_steps}, otherwise it will be called @command{integrate_adaptive}. For this last possibility the user can set the tolerance for the timestep computation by setting a value to the option @var{"Tau"}, that as default value has @math{1.e-6}.
%# 
%# The third input argument must contain the initial value for the unknown. This must contain the generalized coordinates in its first half and the generalized momenta in the second half. Each momentum correspond to the coordinate with the same relative position inside the vector of initial conditions.
%# 
%# The fourth input argument is not mandatory and it should contain a structure with valid ODE fields.
%# @end deftypefn
%#
%# @seealso{odepkg}

function [varargout] = odeVV (vfun, vslot, vinit, varargin)

  order = 2;
  solver = 'odeVV';

  if (nargin == 0) %# Check number and types of all input arguments
    help (solver);
    error ('OdePkg:InvalidArgument', ...
      'number of input arguments must be greater than zero');
  end

  if (nargin < 3)
    print_usage;
  end
  
  if (nargin >= 4)
    if (~isstruct (varargin{1}))
      %# varargin{1:len} are parameters for vfun
      odeopts = odeset;
      odeopts.vfunarguments = varargin;
    elseif (length (varargin) > 1)
      %# varargin{1} is an OdePkg options structure vopt
      odeopts = odepkg_structure_check (varargin{1}, 'odeVV');
      odeopts.vfunarguments = {varargin{2:length(varargin)}};
    else %# if (isstruct (varargin{1}))
      odeopts = odepkg_structure_check (varargin{1}, 'odeVV');
      odeopts.vfunarguments = {};
    end
  else %# if (nargin == 3)
    odeopts = odeset;
    odeopts.vfunarguments = {};
  end

  if (~isvector (vslot) || ~isnumeric(vslot))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  end

  if(length(vslot)<2 && (isempty(odeopts.TimeStepSize) || isempty(odeopts.TimeStepNumber) ))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  elseif(vslot(2)==vslot(1))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  else
    odeopts.vdirection = sign(vslot(2)-vslot(1));
  end
  vslot = vslot(:);

  if (~isvector (vinit) || ~isnumeric (vinit))
    error ('OdePkg:InvalidArgument', ...
      'third input argument must be a valid numerical value or vector');
  end
  vinit = vinit(:);

  vinit_size = length(vinit);
  if(mod(vinit_size,2) != 0)
    error ('OdePkg:InvalidArgument', ...
      'third input argument must be a vector with an even number of components');
  end

  if ~(isa (vfun, 'function_handle') || isa (vfun, 'inline'))
    error ('OdePkg:InvalidArgument', ...
      'first input argument must be a valid function handle');
  end

  flag = 0;
  try
    vfun(vslot(1),vinit){1};
  catch
    flag = 1;
  end

  if(flag==0)
    vfun = @(t,x)vfun(t,x){1};
  end

  warning('OdePkg:InvalidArgument', ...
    'solver assuming initial conditions vector contains coordinates in first half and velocities in second half');

  %# Start preprocessing, have a look which options are set in
  %# odeopts, check if an invalid or unused option is set
  if(isempty(odeopts.TimeStepNumber) && isempty(odeopts.TimeStepSize))
    integrate_func = 'adaptive';
    odeopts.vstepsizefixed = false;
  elseif(~isempty(odeopts.TimeStepNumber) && ~isempty(odeopts.TimeStepSize))
    integrate_func = 'n_steps';
    odeopts.vstepsizefixed = true;
    if(sign(odeopts.TimeStepSize)!=odeopts.vdirection)
      warning('OdePkg:InvalidArgument', ...
        'option "TimeStepSize" has a wrong sign, it will be corrected automatically');
      odeopts.TimeStepSize = (-1)*odeopts.TimeStepSize;
    end
  elseif(isempty(odeopts.TimeStepNumber) && ~isempty(odeopts.TimeStepSize))
    integrate_func = 'const';
    odeopts.vstepsizefixed = true;
    if(sign(odeopts.TimeStepSize)!=odeopts.vdirection)
      warning('OdePkg:InvalidArgument', ...
        'option "TimeStepSize" has a wrong sign, it will be corrected automatically');
      odeopts.TimeStepSize = (-1)*odeopts.TimeStepSize;
    end
  else
    warning ('OdePkg:InvalidArgument', ...
      'assuming an adaptive integrate function');
    integrate_func = 'adaptive';
  end

  %# Get the default options that can be set with 'odeset' temporarily
  vodetemp = odeset;

  %# Implementation of the option RelTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (odeopts.RelTol) && ~odeopts.vstepsizefixed)
    odeopts.RelTol = 1e-3;
    warning ('OdePkg:InvalidArgument', ...
      'Option "RelTol" not set, new value %f is used', odeopts.RelTol);
  elseif (~isempty (odeopts.RelTol) && odeopts.vstepsizefixed)
    warning ('OdePkg:InvalidArgument', ...
      'Option "RelTol" will be ignored if fixed time stamps are given');
  end

  %# Implementation of the option AbsTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (odeopts.AbsTol) && ~odeopts.vstepsizefixed)
    odeopts.AbsTol = 1e-6;
    warning ('OdePkg:InvalidArgument', ...
      'Option "AbsTol" not set, new value %f is used', odeopts.AbsTol);
  elseif (~isempty (odeopts.AbsTol) && odeopts.vstepsizefixed)
    warning ('OdePkg:InvalidArgument', ...
      'Option "AbsTol" will be ignored if fixed time stamps are given');
  else
    odeopts.AbsTol = odeopts.AbsTol(:); %# Create column vector
  end

  %# Implementation of the option NormControl has been finished. This
  %# option can be set by the user to another value than default value.
  if (strcmp (odeopts.NormControl, 'on'))
    odeopts.normcontrol = true;
  else
    odeopts.normcontrol = false;
  end

  %# Implementation of the option NonNegative has been finished. This
  %# option can be set by the user to another value than default value.
  if (~isempty (odeopts.NonNegative))
    if (isempty (odeopts.Mass))
      odeopts.havenonnegative = true;
    else
      odeopts.havenonnegative = false;
      warning ('OdePkg:InvalidArgument', ...
        'Option "NonNegative" will be ignored if mass matrix is set');
    end
  else 
    odeopts.havenonnegative = false;
  end

  %# Implementation of the option OutputFcn has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (odeopts.OutputFcn) && nargout == 0)
    odeopts.OutputFcn = @odeplot;
    odeopts.haveoutputfunction = true;
  elseif (isempty (odeopts.OutputFcn))
    odeopts.haveoutputfunction = false;
  else
    odeopts.haveoutputfunction = true;
  end

  %# Implementation of the option OutputSel has been finished. This
  %# option can be set by the user to another value than default value.
  if (~isempty (odeopts.OutputSel))
    odeopts.vhaveoutputselection = true;
  else
    odeopts.vhaveoutputselection = false;
  end

  %# Implementation of the option Refine has been finished. This option
  %# can be set by the user to another value than default value.
  if (odeopts.Refine > 0)
    odeopts.vhaverefine = true;
  else
    odeopts.vhaverefine = false;
  end

  %# Implementation of the option InitialStep has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (odeopts.InitialStep) && strcmp(integrate_func,'adaptive'))
    odeopts.InitialStep = odeopts.vdirection* ...
      starting_stepsize (order, vfun, vslot(1), vinit, odeopts.AbsTol, odeopts.RelTol, odeopts.normcontrol);
    warning ('OdePkg:InvalidArgument', ...
      'option "InitialStep" not set, estimated value %f is used', odeopts.InitialStep);
  elseif(isempty (odeopts.InitialStep))
    odeopts.InitialStep = odeget(odeopts,'TimeStepSize');
  end

  %# Implementation of the option MaxStep has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (odeopts.MaxStep) && ~odeopts.vstepsizefixed)
    odeopts.MaxStep = abs (vslot(2) - vslot(1)) / 10;
    warning ('OdePkg:InvalidArgument', ...
      'Option "MaxStep" not set, new value %f is used', odeopts.MaxStep);
  end

  %# Implementation of the option Events has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isempty (odeopts.Events))
    odeopts.haveeventfunction = true;
  else
    odeopts.haveeventfunction = false;
  end

  %# The options 'Jacobian', 'JPattern' and 'Vectorized' will be ignored
  %# by this solver because this solver uses an explicit Runge-Kutta
  %# method and therefore no Jacobian calculation is necessary
  if (~isequal (odeopts.Jacobian, vodetemp.Jacobian))
    warning ('OdePkg:InvalidArgument', ...
      'option "Jacobian" will be ignored by this solver');
  end

  if (~isequal (odeopts.JPattern, vodetemp.JPattern))
    warning ('OdePkg:InvalidArgument', ...
      'option "JPattern" will be ignored by this solver');
  end

  if (~isequal (odeopts.Vectorized, vodetemp.Vectorized))
    warning ('OdePkg:InvalidArgument', ...
      'option "Vectorized" will be ignored, not yet implemented functionality');
  end

  if (! isfield (odeopts, "MaxNewtonIterations") || isempty (odeopts.MaxNewtonIterations))
    odeopts.MaxNewtonIterations = 100;
    warning ("OdePkg:InvalidArgument",
             "odeVV: option 'MaxNewtonIterations' not set, default value %i is used",
             odeopts.MaxNewtonIterations);
  endif

  if (! isfield (odeopts, "NewtonTol") || isempty (odeopts.NewtonTol))
    odeopts.NewtonTol = 1e-7;
    warning ("OdePkg:InvalidArgument",
             "odeVV: option 'NewtonTol' not set, default value %g is used",
             odeopts.NewtonTol);
  endif

  if (! isfield (odeopts, "UseJacobian") || isempty (odeopts.UseJacobian))
    odeopts.UseJacobian = "no";
    warning ("OdePkg:InvalidArgument",
             "odeVV: option 'UseJacobian' not set, default value '%s' is used",
             odeopts.UseJacobian);
  endif

  %# Implementation of the option Mass has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isequal (odeopts.Mass, vodetemp.Mass))
    warning ('OdePkg:InvalidArgument', ...
      'option "Mass" will be ignored by this solver');
  end

  %# Implementation of the option MStateDependence has been finished.
  %# This option can be set by the user to another value than default
  %# value. 
  if (~isequal (odeopts.MStateDependence, vodetemp.MStateDependence))
    warning ('OdePkg:InvalidArgument', ...
      'option "MStateDependence" will be ignored by this solver');
  end

  %# Other options that are not used by this solver. Print a warning
  %# message to tell the user that the option(s) is/are ignored.
  if (~isequal (odeopts.MvPattern, vodetemp.MvPattern))
    warning ('OdePkg:InvalidArgument', ...
      'option "MvPattern" will be ignored by this solver');
  end

  if (~isequal (odeopts.MassSingular, vodetemp.MassSingular))
    warning ('OdePkg:InvalidArgument', ...
      'option "MassSingular" will be ignored by this solver');
  end

  if (~isequal (odeopts.InitialSlope, vodetemp.InitialSlope))
    warning ('OdePkg:InvalidArgument', ...
      'option "InitialSlope" will be ignored by this solver');
  end

  if (~isequal (odeopts.MaxOrder, vodetemp.MaxOrder))
    warning ('OdePkg:InvalidArgument', ...
      'option "MaxOrder" will be ignored by this solver');
  end

  if (~isequal (odeopts.BDF, vodetemp.BDF))
    warning ('OdePkg:InvalidArgument', ...
      'option "BDF" will be ignored by this solver');
  end

  %# Starting the initialisation of the core solver odeVV
  SubOpts = odeopts;

  switch integrate_func
    case 'adaptive'
      solution = integrate_adaptive(@velocity_verlet,order,vfun,vslot,vinit,SubOpts);
    case 'n_steps'
      solution = integrate_n_steps(@velocity_verlet,vfun,vslot(1),vinit,odeopts.TimeStepSize,odeopts.TimeStepNumber,SubOpts);
    case 'const'
      solution = integrate_const(@velocity_verlet,vfun,vslot,vinit,odeopts.TimeStepSize,SubOpts);
  end

  %# Postprocessing, do whatever when terminating integration algorithm
  if (odeopts.haveoutputfunction) %# Cleanup plotter
    feval (odeopts.OutputFcn, solution.t(end), ...
      solution.x(end,:)', 'done', odeopts.vfunarguments{:});
  end
  if (odeopts.haveeventfunction)  %# Cleanup event function handling
    odepkg_event_handle (odeopts.Events, solution.t(end), ...
      solution.x(end,:)', 'done', odeopts.vfunarguments{:});
  end

  %# Print additional information if option Stats is set
  if (strcmp (odeopts.Stats, 'on'))
    vhavestats = true;
    vnsteps    = solution.vcntloop-2;                    %# vcntloop from 2..end
    vnfailed   = (solution.vcntcycles-1)-(solution.vcntloop-2)+1; %# vcntcycl from 1..end
    vndecomps  = 0;                             %# number of LU decompositions
    vnpds      = 0;                             %# number of partial derivatives
    vnlinsols  = 0;                             %# no. of solutions of linear systems
    %# Print cost statistics if no output argument is given
    if (nargout == 0)
      vmsg = fprintf (1, 'Number of successful steps: %d\n', vnsteps);
      vmsg = fprintf (1, 'Number of failed attempts:  %d\n', vnfailed);
    end
  else
    vhavestats = false;
  end

  if (nargout == 1)                 %# Sort output variables, depends on nargout
    varargout{1}.x = solution.t;   %# Time stamps are saved in field x
    varargout{1}.y = solution.x; %# Results are saved in field y
    varargout{1}.solver = solver;  %# Solver name is saved in field solver
    if (odeopts.haveeventfunction)
      varargout{1}.ie = solution.vevent{2};  %# Index info which event occured
      varargout{1}.xe = solution.vevent{3};  %# Time info when an event occured
      varargout{1}.ye = solution.vevent{4};  %# Results when an event occured
    end
    if (vhavestats)
      varargout{1}.stats = struct;
      varargout{1}.stats.nsteps   = vnsteps;
      varargout{1}.stats.nfailed  = vnfailed;
      varargout{1}.stats.npds     = vnpds;
      varargout{1}.stats.ndecomps = vndecomps;
      varargout{1}.stats.nlinsols = vnlinsols;
    end
  elseif (nargout == 2)
    varargout{1} = solution.t;     %# Time stamps are first output argument
    varargout{2} = solution.x;   %# Results are second output argument
  elseif (nargout == 5)
    varargout{1} = solution.t;     %# Same as (nargout == 2)
    varargout{2} = solution.x;   %# Same as (nargout == 2)
    varargout{3} = [];              %# LabMat doesn't accept lines like
    varargout{4} = [];              %# varargout{3} = varargout{4} = [];
    varargout{5} = [];
    if (odeopts.haveeventfunction)
      varargout{3} = solution.vevent{3};     %# Time info when an event occured
      varargout{4} = solution.vevent{4};     %# Results when an event occured
      varargout{5} = solution.vevent{2};     %# Index info which event occured
    end
  end

end

%! # We are using the "armonic oscillator" implementation for all tests that
%! # are done for this function.
%!function [ydot] = armonic (vt, vy)
%!  omega = 1.0;
%!  ydot = [vy(length(vy)/2+1:end); -(omega^2)*vy(1:length(vy)/2)];
%!endfunction
%!
%! %# Turn off output of warning messages for all tests, turn them on
%! %# again if the last test is called
%!error %# output argument
%!  warning ('off', 'OdePkg:InvalidArgument');
%!  B = odeVV (1, [1 5], [1 0]);
%!error %# input argument number one
%!  [vt, vy] = odeVV (1, [1 5], [1 0]);
%!error %# input argument number two
%!  [vt, vy] = odeVV (@armonic, 1, [1 0]);
%!error %# input argument number three
%!  [vt, vy] = odeVV (@armonic, [1 5], [1 0 0]);
%!test %# two output arguments
%!  [vt, vy] = odeVV (@armonic, [0 1], [1 0]);
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# extra input arguments passed through
%!  [vt, vy] = odeVV (@armonic, [0 1], [1 0], 12, 13, 'KL');
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# empty OdePkg structure *but* extra input arguments
%!  vopt = odeset;
%!  [vt, vy] = odeVV (@armonic, [0 1], [1 0], vopt, 12, 13, 'KL');
%!  assert ([vt(end), vy(end,1)], [1, cos(vt(end))], 1e-2);
%!test %# InitialStep option
%!  vopt = odeset ('InitialStep', 1e-8);
%!  [vt, vy] = odeVV (@armonic, [0 0.2], [1 0], vopt);
%!  assert ([vt(2)-vt(1)], [1e-8], 1e-9);
%!
%!  warning ('on', 'OdePkg:InvalidArgument');

%# Local Variables: ***
%# mode: octave ***
%# End: ***
