%# Copyright (C) 2006-2012, Thomas Treichl <treichl@users.sourceforge.net>
%# Copyright (C) 2009-2012, Sebastian Schoeps <schoeps@math.uni-wuppertal.de>
%# Copyright (C) 2013, Roberto Porcu' <roberto.porcu@polimi.it>
%# OdePkg - A package for solving ordinary differential equations and more
%#
%# This program is free software; you can redistribute it and/or modify
%# it under the terms of the GNU General Public License as published by
%# the Free Software Foundation; either version 2 of the License, or
%# (at your option) any later version.
%#
%# This program is distributed in the hope that it will be useful,
%# but WITHOUT ANY WARRANTY; without even the implied warranty of
%# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%# GNU General Public License for more details.
%#
%# You should have received a copy of the GNU General Public License
%# along with this program; If not, see <http://www.gnu.org/licenses/>.

%# -*- texinfo -*-
%# @deftypefn {Command} {[@var{sol}] =} odebwe (@var{@@fun}, @var{slot}, @var{init}, [@var{opt}], [@var{par1}, @var{par2}, @dots{}])
%# @deftypefnx {Command} {[@var{t}, @var{y}, [@var{xe}, @var{ye}, @var{ie}]] =} odebwe (@var{@@fun}, @var{slot}, @var{init}, [@var{opt}], [@var{par1}, @var{par2}, @dots{}])
%# 
%# This function file can be used to solve a set of non--stiff ordinary differential equations (non--stiff ODEs) with the well known implicit Backward-Euler method of order 1.
%# 
%# This function can be called with two output arguments: @var{t} and @var{y}. Variable @var{t} is a column vector and contains the time stamps, instead @var{y} is a matrix in which each column refers to a different unknown of the problem and the rows number is the same of @var{t} rows number so that each row of @var{y} contains the values of all unknowns at the time value contained in the corresponding row in @var{t}.
%# 
%# The first input argument must be a function_handle or an inline function that defines the set of ODE: @code{y' = f(t,y)}.
%#  
%# As described above, this function must take two input arguments, where the first is the time and the second the unknowns, and must have just one output argument that should be a vector.
%# 
%# The second input argument must contain time informations. Usually it should be a vector with at least two elements which define the initial and the final time instants; if the elements are more than two, then the solution will be evaluated also at these intermediate time instants unless the integrate function called is the @command{integrate_n_steps}. If there is only one time value, then it will give an error unless the options structure has no empty fields named @var{"TimeStepNumber"} and @var{"TimeStepSize"}. If the option @var{"TimeStepSize"} is not empty, then the stepper called will be @command{integrate_const}, if also @var{"TimeStepNumber"} is not empty it will be called the integrate function @command{integrate_n_steps}, otherwise it will be called @command{integrate_adaptive}. For this last possibility the user can set the tolerance for the timestep computation by setting a value to the option @var{"Tau"}, that as default value has @math{1.e-6}.
%# 
%# The third input argument must contain the initial value for the unknown. If this is a vector then the solution @var{y} will be a matrix in which each column is the solution for the corresponding initial value in @var{init}.
%# 
%# The fourth input argument is not mandatory and it should contain a structure with valid ODE options. The option @var{"InexactSolver"} can be set to @var{'inexact_newton'} or @var{'newton_raphson'}. If is set to @var{'newton_raphson'} it will be used this stepper. If is not set to @var{'newton_raphson'} it will be used the stepper@command{inexact_newton} and in this case if it is set to @var{'inexact_newton'} this inexact solver will be used in the stepper, otherwiseit will be used @command{fsolve}. The Jacobian function can be passed in the option @var{"Jacobian"} as a function_handle and to use it the option @var{"UseJacobian"} must be set on @var{'yes'}.If no Jacobian is given but is requested to use it, it will be approximated with the utility function @command{approx_Jacobian}. The @command{inexact_newton} solver has some options that can be set by set them in this fourth argument, they will be explained in the relative section.
%# 
%# For example, solve an anonymous implementation of the Van der Pol equation
%# @example
%# fvdp = @@(t,y) [y(2); (1 - y(1)^2) * y(2) - y(1)];
%# [T,Y] = odebwe (fvdp, [0 20], [2 0]);
%# @end example
%# @end deftypefn
%#
%# @seealso{odepkg}

function [varargout] = odebwe (vfun, vslot, vinit, varargin)

  vorder = 1;
  vsolver = 'odebwe';

  if (nargin == 0) %# Check number and types of all input arguments
    help (vsolver);
    error ('OdePkg:InvalidArgument', ...
      'number of input arguments must be greater than zero');
  end

  if (nargin < 3)
    print_usage;
  end
  
  if (nargin >= 4)
    if (~isstruct (varargin{1}))
      %# varargin{1:len} are parameters for vfun
      vodeoptions = odeset;
      vodeoptions.vfunarguments = varargin;
    elseif (length (varargin) > 1)
      %# varargin{1} is an OdePkg options structure vopt
      vodeoptions = odepkg_structure_check (varargin{1}, 'odebwe');
      vodeoptions.vfunarguments = {varargin{2:length(varargin)}};
    else %# if (isstruct (varargin{1}))
      vodeoptions = odepkg_structure_check (varargin{1}, 'odebwe');
      vodeoptions.vfunarguments = {};
    end
  else %# if (nargin == 3)
    vodeoptions = odeset; 
    vodeoptions.vfunarguments = {};
  end

  if (~isvector (vslot) || ~isnumeric(vslot))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  end

  if(length(vslot)<2 && (isempty(vodeoptions.TimeStepSize) || isempty(vodeoptions.TimeStepNumber) ))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  elseif(vslot(2)==vslot(1))
    error ('OdePkg:InvalidArgument', ...
      'second input argument must be a valid vector');
  else
    vodeoptions.vdirection = sign(vslot(2)-vslot(1));
  end
  vslot = vslot(:);

  if (~isvector (vinit) || ~isnumeric (vinit))
    error ('OdePkg:InvalidArgument', ...
      'third input argument must be a valid numerical value');
  end
  vinit = vinit(:);

  if ~(isa (vfun, 'function_handle') || isa (vfun, 'inline'))
    error ('OdePkg:InvalidArgument', ...
      'first input argument must be a valid function handle');
  end

  %# Start preprocessing, have a look which options are set in
  %# vodeoptions, check if an invalid or unused option is set
  if(isempty(vodeoptions.TimeStepNumber) && isempty(vodeoptions.TimeStepSize))
    integrate_func = 'adaptive';
    vodeoptions.vstepsizefixed = false;
  elseif(~isempty(vodeoptions.TimeStepNumber) && ~isempty(vodeoptions.TimeStepSize))
    integrate_func = 'n_steps';
    vodeoptions.vstepsizefixed = true;
    if(sign(vodeoptions.TimeStepSize)!=vodeoptions.vdirection)
      warning('OdePkg:InvalidArgument', ...
        'option "TimeStepSize" has a wrong sign, it will be corrected automatically');
      vodeoptions.TimeStepSize = (-1)*vodeoptions.TimeStepSize;
    end
  elseif(isempty(vodeoptions.TimeStepNumber) && ~isempty(vodeoptions.TimeStepSize))
    integrate_func = 'const';
    vodeoptions.vstepsizefixed = true;
    if(sign(vodeoptions.TimeStepSize)!=vodeoptions.vdirection)
      warning('OdePkg:InvalidArgument', ...
        'option "TimeStepSize" has a wrong sign, it will be corrected automatically');
      vodeoptions.TimeStepSize = (-1)*vodeoptions.TimeStepSize;
    end
  else
    warning ('OdePkg:InvalidArgument', ...
      'assuming an adaptive integrate function');
    integrate_func = 'adaptive';
  end

  %# Get the default options that can be set with 'odeset' temporarily
  vodetemp = odeset;

  %# Implementation of the option RelTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.RelTol) && ~vodeoptions.vstepsizefixed)
    vodeoptions.RelTol = 1e-3;
    warning ('OdePkg:InvalidArgument', ...
      'Option "RelTol" not set, new value %f is used', vodeoptions.RelTol);
  elseif (~isempty (vodeoptions.RelTol) && vodeoptions.vstepsizefixed)
    warning ('OdePkg:InvalidArgument', ...
      'Option "RelTol" will be ignored if fixed time stamps are given');
  end

  %# Implementation of the option AbsTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.AbsTol) && ~vodeoptions.vstepsizefixed)
    vodeoptions.AbsTol = 1e-6;
    warning ('OdePkg:InvalidArgument', ...
      'Option "AbsTol" not set, new value %f is used', vodeoptions.AbsTol);
  elseif (~isempty (vodeoptions.AbsTol) && vodeoptions.vstepsizefixed)
    warning ('OdePkg:InvalidArgument', ...
      'Option "AbsTol" will be ignored if fixed time stamps are given');
  else
    vodeoptions.AbsTol = vodeoptions.AbsTol(:); %# Create column vector
  end

  %# Implementation of the option NormControl has been finished. This
  %# option can be set by the user to another value than default value.
  if (strcmp (vodeoptions.NormControl, 'on')) 
    vodeoptions.vnormcontrol = true;
  else 
    vodeoptions.vnormcontrol = false; 
  end

  %# Implementation of the option NonNegative has been finished. This
  %# option can be set by the user to another value than default value.
  if (~isempty (vodeoptions.NonNegative))
    if (isempty (vodeoptions.Mass))
      vodeoptions.vhavenonnegative = true;
    else
      vodeoptions.vhavenonnegative = false;
      warning ('OdePkg:InvalidArgument', ...
        'Option "NonNegative" will be ignored if mass matrix is set');
    end
  else 
    vodeoptions.vhavenonnegative = false;
  end

  %# Implementation of the option OutputFcn has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (vodeoptions.OutputFcn) && nargout == 0)
    vodeoptions.OutputFcn = @odeplot;
    vodeoptions.vhaveoutputfunction = true;
  elseif (isempty (vodeoptions.OutputFcn))
    vodeoptions.vhaveoutputfunction = false;
  else 
    vodeoptions.vhaveoutputfunction = true;
  end

  %# Implementation of the option OutputSel has been finished. This
  %# option can be set by the user to another value than default value.
  if (~isempty (vodeoptions.OutputSel))
    vodeoptions.vhaveoutputselection = true;
  else 
    vodeoptions.vhaveoutputselection = false; 
  end

  %# Implementation of the option OutputSave has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (vodeoptions.OutputSave))
    vodeoptions.OutputSave = 1;
  end

  %# Implementation of the option Refine has been finished. This option
  %# can be set by the user to another value than default value.
  if (vodeoptions.Refine > 0)
    vodeoptions.vhaverefine = true;
  else 
    vodeoptions.vhaverefine = false;
  end

  %# Implementation of the option Stats has been finished. This option
  %# can be set by the user to another value than default value.

  %# Implementation of the option InitialStep has been finished. This
  %# option can be set by the user to another value than default value.
  if (isempty (vodeoptions.InitialStep) && strcmp(integrate_func,'adaptive'))
    vodeoptions.InitialStep = vodeoptions.vdirection* ...
      starting_stepsize (vorder, vfun, vslot(1), vinit, vodeoptions.AbsTol, vodeoptions.RelTol, vodeoptions.vnormcontrol);
    warning ('OdePkg:InvalidArgument', ...
      'option "InitialStep" not set, estimated value %f is used', vodeoptions.InitialStep);
  elseif(isempty (vodeoptions.InitialStep))
    vodeoptions.InitialStep = odeget(vodeoptions,'TimeStepSize');
  end

  %# Implementation of the option MaxNewtonIterations has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.MaxNewtonIterations))
    vodeoptions.MaxNewtonIterations = 100;
    warning ('OdePkg:InvalidArgument', ...
      'option "MaxNewtonIterations" not set, default value %d is used', vodeoptions.MaxNewtonIterations);
  end

  %# Implementation of the option NewtonTol has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.NewtonTol))
    vodeoptions.NewtonTol = 1e-7;
    warning ('OdePkg:InvalidArgument', ...
      'option "NewtonTol" not set, default value %g is used', vodeoptions.NewtonTol);
  end

  %# Implementation of the option MaxStep has been finished. This option
  %# can be set by the user to another value than default value.
  if (isempty (vodeoptions.MaxStep) && ~vodeoptions.vstepsizefixed)
    vodeoptions.MaxStep = abs (vslot(2) - vslot(1)) / 10;
    warning ('OdePkg:InvalidArgument', ...
      'Option "MaxStep" not set, new value %f is used', vodeoptions.MaxStep);
  end

  %# Implementation of the option Events has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isempty (vodeoptions.Events))
    vodeoptions.vhaveeventfunction = true;
  else 
    vodeoptions.vhaveeventfunction = false;
  end

  %# Implementation of the option Jacobian has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isempty (vodeoptions.Jacobian) && isnumeric (vodeoptions.Jacobian))
    vhavejachandle = false;
    visjacapproximated = false;
  elseif (isa (vodeoptions.Jacobian, 'function_handle'))
    vhavejachandle = true; %# jac defined by a function handle
    visjacapproximated = false;
  else %# no Jacobian - we will use numerical differentiation
    vhavejachandle = true;
    visjacapproximated = true;
    vodeoptions.Jacobian = @(t,x)approx_Jacobian(vfun,t,x);
  end
  
  %# Implementation of the option Mass has been finished. This option
  %# can be set by the user to another value than default value.
  if (~isempty (vodeoptions.Mass) && isnumeric (vodeoptions.Mass))
    vhavemasshandle = false;
  elseif (isa (vodeoptions.Mass, 'function_handle'))
    vhavemasshandle = true; %# mass defined by a function handle
  else %# no mass matrix - creating a diag-matrix of ones for mass
    vhavemasshandle = false; 
    vodeoptions.Mass = sparse (eye (length (vinit)) );
  end

  %# Implementation of the option MStateDependence has been finished.
  %# This option can be set by the user to another value than default
  %# value. 
  if (strcmp (vodeoptions.MStateDependence, 'none'))
    vmassdependence = false;
  else 
    vmassdependence = true;
  end

  %# Other options that are not used by this solver. Print a warning
  %# message to tell the user that the option(s) is/are ignored.
  if (~isequal (vodeoptions.NonNegative, vodetemp.NonNegative))
    warning ('OdePkg:InvalidArgument', ...
      'option "NonNegative" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.JPattern, vodetemp.JPattern))
    warning ('OdePkg:InvalidArgument', ...
      'option "JPattern" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.Vectorized, vodetemp.Vectorized))
    warning ('OdePkg:InvalidArgument', ...
      'option "Vectorized" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.MvPattern, vodetemp.MvPattern))
    warning ('OdePkg:InvalidArgument', ...
      'option "MvPattern" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.MassSingular, vodetemp.MassSingular))
    warning ('OdePkg:InvalidArgument', ...
      'option "MassSingular" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.InitialSlope, vodetemp.InitialSlope))
    warning ('OdePkg:InvalidArgument', ...
      'option "InitialSlope" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.MaxOrder, vodetemp.MaxOrder))
    warning ('OdePkg:InvalidArgument', ...
      'option "MaxOrder" will be ignored by this solver');
  end

  if (~isequal (vodeoptions.BDF, vodetemp.BDF))
    warning ('OdePkg:InvalidArgument', ...
      'option "BDF" will be ignored by this solver');
  end

  %# the following option enables the simplified Newton method
  %# which evaluates the Jacobian only once instead of the
  %# standard method that updates the Jacobian in each iteration
  vsimplified = false; % or true

  %# Starting the core solver odebwe
  SubOpts = vodeoptions;

  SubOpts.havejachandle = vhavejachandle;
  SubOpts.havemasshandle = vhavemasshandle;
  SubOpts.massdependence = vmassdependence;
  SubOpts.issimplified = vsimplified;

  inexact_solver = odeget(vodeoptions,'InexactSolver','newton_raphson','fast_not_empty');
  inexact_solver = lower(deblank(inexact_solver));
  SubOpts.InexactSolver = inexact_solver;
  
  if( strcmp(inexact_solver,'newton_raphson') )

    if(visjacapproximated)
      warning('OdePkg:InvalidArgument', ...
        'missing function''s Jacobian in options. It will be approximated');
    end

    switch integrate_func
      case 'adaptive'
        SobOpts.Tau = odeget(vodeoptions,'Tau',1.e-6,'fast_not_empty');
        solution = integrate_adaptive(@bwe_newton_raphson,vorder,vfun,vslot,vinit,SubOpts);
      case 'n_steps'
        solution = integrate_n_steps(@bwe_newton_raphson,vfun,vslot(1),vinit,vodeoptions.TimeStepSize,vodeoptions.TimeStepNumber,SubOpts);
      case 'const'
        solution = integrate_const(@bwe_newton_raphson,vfun,vslot,vinit,vodeoptions.TimeStepSize,SubOpts);
    end

  else

    use_jacob = odeget(vodeoptions,'UseJacobian','no','fast_not_empty');
    use_jacob = lower(deblank(use_jacob));
    SubOpts.UseJacobian = use_jacob;

    if( visjacapproximated && strcmp(use_jacob,'yes') )
      warning('OdePkg:InvalidArgument', ...
        'missing function''s Jacobian in options. It will be approximated');
    end

    switch integrate_func
      case 'adaptive'
        solution = integrate_adaptive(@bwe_inexact_newton,vorder,vfun,vslot,vinit,SubOpts);
      case 'n_steps'
        solution = integrate_n_steps(@bwe_inexact_newton,vfun,vslot(1),vinit,vodeoptions.TimeStepSize,vodeoptions.TimeStepNumber,SubOpts);
      case 'const'
        solution = integrate_const(@bwe_inexact_newton,vfun,vslot,vinit,vodeoptions.TimeStepSize,SubOpts);
    end
  end

  %# Postprocessing, do whatever when terminating integration algorithm
  if (vodeoptions.vhaveoutputfunction) %# Cleanup plotter
    feval (vodeoptions.OutputFcn, solution.t(end), ...
      solution.x(end,:)', 'done', vodeoptions.vfunarguments{:});
  end
  if (vodeoptions.vhaveeventfunction)  %# Cleanup event function handling
    odepkg_event_handle (vodeoptions.Events, solution.t(end), ...
      solution.x(end,:)', 'done', vodeoptions.vfunarguments{:});
  end

  %# Print additional information if option Stats is set
  if (strcmp (vodeoptions.Stats, 'on'))
    vhavestats = true;
    vnsteps    = solution.vcntloop-2;                    %# vcntloop from 2..end
    vnfailed   = (solution.vcntcycles-1)-(solution.vcntloop-2)+1; %# vcntcycl from 1..end
    vndecomps  = 0;                             %# number of LU decompositions
    vnpds      = 0;                             %# number of partial derivatives
    vnlinsols  = 0;                             %# no. of solutions of linear systems
    %# Print cost statistics if no output argument is given
    if (nargout == 0)
      vmsg = fprintf (1, 'Number of successful steps: %d\n', vnsteps);
      vmsg = fprintf (1, 'Number of failed attempts:  %d\n', vnfailed);
    end
  else
    vhavestats = false;
  end

  if (nargout == 1)                 %# Sort output variables, depends on nargout
    varargout{1}.x = solution.t;   %# Time stamps are saved in field x
    varargout{1}.y = solution.x; %# Results are saved in field y
    varargout{1}.solver = vsolver;  %# Solver name is saved in field solver
    if (vodeoptions.vhaveeventfunction) 
      varargout{1}.ie = solution.vevent{2};  %# Index info which event occured
      varargout{1}.xe = solution.vevent{3};  %# Time info when an event occured
      varargout{1}.ye = solution.vevent{4};  %# Results when an event occured
    end
    if (vhavestats)
      varargout{1}.stats = struct;
      varargout{1}.stats.nsteps   = vnsteps;
      varargout{1}.stats.nfailed  = vnfailed;
      varargout{1}.stats.npds     = vnpds;
      varargout{1}.stats.ndecomps = vndecomps;
      varargout{1}.stats.nlinsols = vnlinsols;
    end
  elseif (nargout == 2)
    varargout{1} = solution.t;     %# Time stamps are first output argument
    varargout{2} = solution.x;   %# Results are second output argument
  elseif (nargout == 5)
    varargout{1} = solution.t;     %# Same as (nargout == 2)
    varargout{2} = solution.x;   %# Same as (nargout == 2)
    varargout{3} = [];              %# LabMat doesn't accept lines like
    varargout{4} = [];              %# varargout{3} = varargout{4} = [];
    varargout{5} = [];
    if (vodeoptions.vhaveeventfunction) 
      varargout{3} = solution.vevent{3};     %# Time info when an event occured
      varargout{4} = solution.vevent{4};     %# Results when an event occured
      varargout{5} = solution.vevent{2};     %# Index info which event occured
    end
  end

end

%! # We are using the "Van der Pol" implementation for all tests that
%! # are done for this function. We also define a Jacobian, Events,
%! # pseudo-Mass implementation. For further tests we also define a
%! # reference solution (computed at high accuracy) and an OutputFcn
%!function [ydot] = fpol (vt, vy, varargin) %# The Van der Pol
%!  ydot = [vy(2); (1 - vy(1)^2) * vy(2) - vy(1)];
%!function [vjac] = fjac (vt, vy, varargin) %# its Jacobian
%!  vjac = [0, 1; -1 - 2 * vy(1) * vy(2), 1 - vy(1)^2];
%!function [vjac] = fjcc (vt, vy, varargin) %# sparse type
%!  vjac = sparse ([0, 1; -1 - 2 * vy(1) * vy(2), 1 - vy(1)^2]);
%!function [vval, vtrm, vdir] = feve (vt, vy, varargin)
%!  vval = fpol (vt, vy, varargin); %# We use the derivatives
%!  vtrm = zeros (2,1);             %# that's why component 2
%!  vdir = ones (2,1);              %# seems to not be exact
%!function [vval, vtrm, vdir] = fevn (vt, vy, varargin)
%!  vval = fpol (vt, vy, varargin); %# We use the derivatives
%!  vtrm = ones (2,1);              %# that's why component 2
%!  vdir = ones (2,1);              %# seems to not be exact
%!function [vmas] = fmas (vt, vy)
%!  vmas = [1, 0; 0, 1];            %# Dummy mass matrix for tests
%!function [vmas] = fmsa (vt, vy)
%!  vmas = sparse ([1, 0; 0, 1]);   %# A sparse dummy matrix
%!function [vref] = fref ()         %# The computed reference sol
%!  vref = [0.32331666704577, -1.83297456798624];
%!function [vout] = fout (vt, vy, vflag, varargin)
%!  if (regexp (char (vflag), 'init') == 1)
%!    if (any (size (vt) ~= [2, 1])) error ('"fout" step "init"'); end
%!  elseif (isempty (vflag))
%!    if (any (size (vt) ~= [1, 1])) error ('"fout" step "calc"'); end
%!    vout = false;
%!  elseif (regexp (char (vflag), 'done') == 1)
%!    if (any (size (vt) ~= [1, 1])) error ('"fout" step "done"'); end
%!  else error ('"fout" invalid vflag');
%!  end
%!
%! %# Turn off output of warning messages for all tests, turn them on
%! %# again if the last test is called
%!error %# input argument number one
%!  warning ('off', 'OdePkg:InvalidArgument');
%!  B = odebwe (1, [0 25], [3 15 1]);
%!error %# input argument number two
%!  B = odebwe (@fpol, 1, [3 15 1]);
%!error %# input argument number three
%!  B = odebwe (@flor, [0 25], 1);
%!test %# one output argument
%!  vsol = odebwe (@fpol, [0 2], [2 0]);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!  assert (isfield (vsol, 'solver'));
%!  assert (vsol.solver, 'odebwe');
%!test %# two output arguments
%!  [vt, vy] = odebwe (@fpol, [0 2], [2 0]);
%!  assert ([vt(end), vy(end,:)], [2, fref], 1e-2);
%!test %# five output arguments and no Events
%!  [vt, vy, vxe, vye, vie] = odebwe (@fpol, [0 2], [2 0]);
%!  assert ([vt(end), vy(end,:)], [2, fref], 1e-2);
%!  assert ([vie, vxe, vye], []);
%!test %# anonymous function instead of real function
%!  fvdb = @(vt,vy) [vy(2); (1 - vy(1)^2) * vy(2) - vy(1)];
%!  vsol = odebwe (fvdb, [0 2], [2 0]);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# extra input arguments passed trhough
%!  vsol = odebwe (@fpol, [0 2], [2 0], 12, 13, 'KL');
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# empty OdePkg structure *but* extra input arguments
%!  vopt = odeset;
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt, 12, 13, 'KL');
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!error %# strange OdePkg structure
%!  vopt = struct ('foo', 1);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!test %# Solve vdp in fixed step sizes
%!  vopt = odeset ('TimeStepSize', 0.001);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert (vsol.x(:), [0:0.001:2]');
%!  assert (vsol.y(end,:), fref, 1e-2);
%!test %# Solve in backward direction starting at t=0
%! %# vref = [-1.2054034414, 0.9514292694];
%!  vsol = odebwe (@fpol, [0 -2], [2 0]);
%! %# assert ([vsol.x(end), vsol.y(end,:)], [-2, fref], 1e-2);
%!test %# Solve in backward direction starting at t=2
%! %#  vref = [-1.2154183302, 0.9433018000];
%!  vsol = odebwe (@fpol, [2 -2], [0.3233166627 -1.8329746843]);
%! %#  assert ([vsol.x(end), vsol.y(end,:)], [-2, fref], 1e-2);
%!test %# AbsTol option
%!  vopt = odeset ('AbsTol', 1e-5);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# AbsTol and RelTol option
%!  vopt = odeset ('AbsTol', 1e-6, 'RelTol', 1e-6);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# RelTol and NormControl option -- higher accuracy
%!  vopt = odeset ('RelTol', 1e-8, 'NormControl', 'on');
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-3);
%!
%! %# test for NonNegative option is missing
%! %# test for OutputSel and Refine option is missing
%!
%!test %# Details of OutputSave can't be tested
%!  vopt = odeset ('OutputSave', 1, 'OutputSel', 1);
%!  vsla = odebwe (@fpol, [0 2], [2 0], vopt);
%!  vopt = odeset ('OutputSave', 2);
%!  vslb = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert (length (vsla.x) > length (vslb.x))
%!test %# Stats must add further elements in vsol
%!  vopt = odeset ('Stats', 'on');
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert (isfield (vsol, 'stats'));
%!  assert (isfield (vsol.stats, 'nsteps'));
%!test %# InitialStep option
%!  vopt = odeset ('InitialStep', 1e-8);
%!  vsol = odebwe (@fpol, [0 0.2], [2 0], vopt);
%!  assert ([vsol.x(2)-vsol.x(1)], [1e-8], 1e-9);
%!test %# MaxStep option
%!  vopt = odeset ('MaxStep', 5e-4);
%!  vsol = odebwe (@fpol, [0 0.2], [2 0], vopt);
%!  assert ([vsol.x(20)-vsol.x(19)], [5e-4], 1e-8);
%!test %# Events option add further elements in vsol
%!  vopt = odeset ('Events', @feve);
%!  vsol = odebwe (@fpol, [0 10], [2 0], vopt);
%!  assert (isfield (vsol, 'ie'));
%!  assert (vsol.ie, [2; 1; 2; 1]);
%!  assert (isfield (vsol, 'xe'));
%!  assert (isfield (vsol, 'ye'));
%!test %# Events option, now stop integration
%!  vopt = odeset ('Events', @fevn, 'NormControl', 'on');
%!  vsol = odebwe (@fpol, [0 10], [2 0], vopt);
%!  assert ([vsol.ie, vsol.xe, vsol.ye], ...
%!    [2.0, 2.496110, -0.830550, -2.677589], 1e-3);
%!test %# Events option, five output arguments
%!  vopt = odeset ('Events', @fevn, 'NormControl', 'on');
%!  [vt, vy, vxe, vye, vie] = odebwe (@fpol, [0 10], [2 0], vopt);
%!  assert ([vie, vxe, vye], ...
%!    [2.0, 2.496110, -0.830550, -2.677589], 1e-3);
%!test %# Jacobian option
%!  vopt = odeset ('Jacobian', @fjac);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# Jacobian option and sparse return value
%!  vopt = odeset ('Jacobian', @fjcc);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!
%! %# test for JPattern option is missing
%! %# test for Vectorized option is missing
%!
%!test %# Mass option as function
%!  vopt = odeset ('Mass', @fmas);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# Mass option as matrix
%!  vopt = odeset ('Mass', eye (2,2));
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# Mass option as sparse matrix
%!  vopt = odeset ('Mass', sparse (eye (2,2)));
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!test %# Mass option as function and sparse matrix
%!  vopt = odeset ('Mass', @fmsa);
%!  vsol = odebwe (@fpol, [0 2], [2 0], vopt);
%!  assert ([vsol.x(end), vsol.y(end,:)], [2, fref], 1e-2);
%!
%! %# test for MStateDependence option is missing
%! %# test for MvPattern option is missing
%! %# test for InitialSlope option is missing
%! %# test for MaxOrder option is missing
%! %# test for BDF option is missing
%!
%!  warning ('on', 'OdePkg:InvalidArgument');




%# Local Variables: ***
%# mode: octave ***
%# End: ***
