%# Copyright (C) 2013, Carlo de Falco <carlo.defalco@polimi.it>
%# Copyright (C) 2013, Roberto Porcu' <roberto.porcu@polimi.it>
%# OdePkg - A package for solving ordinary differential equations and more
%#
%# This program is free software; you can redistribute it and/or modify
%# it under the terms of the GNU General Public License as published by
%# the Free Software Foundation; either version 2 of the License, or
%# (at your option) any later version.
%#
%# This program is distributed in the hope that it will be useful,
%# but WITHOUT ANY WARRANTY; without even the implied warranty of
%# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%# GNU General Public License for more details.
%#
%# You should have received a copy of the GNU General Public License
%# along with this program; If not, see <http://www.gnu.org/licenses/>.

%# -*- texinfo -*-
%# @deftypefn {Command} {[@var{t_next}, @var{x_next}] =} velocity_verlet (@var{@@fun}, @var{t}, @var{x}, @var{dt}, [@var{options}])
%# @deftypefnx {Command} {[@var{t_next}, @var{x_next}, @var{x_est}] =} velocity_verlet (@var{@@fun}, @var{t}, @var{x}, @var{dt}, [@var{options}])
%# 
%# This function can be used to integrate Newton's equations of motion
%# with a given initial condition @var{x} from @var{t} to @var{t+dt},
%# with the Velocity-Verlet method using the Richardson extrapolation
%# method to estimate the error. For details see
%# @url{http://en.wikipedia.org/wiki/Verlet_integration}.
%#
%# @itemize @minus
%# @item The first output argument is the final integration time value. 
%# @item The second output parameter is the computed solution at
%# time @var{t_next}. 
%# @item The third output parameter is a higher order solution for the
%# estimation of the error. 
%# @end itemize
%#
%# @itemize @minus
%# @item The first input argument is the function describing the
%# system of ODEs 
%# to be integrated. 
%# @item The second input parameter is the first extreme of
%# integration interval. 
%# @item The third input argument is the initial condition of the
%# system. 
%# @item The fourth input argument is the timestep, that is the
%# length of the integration interval. 
%# @item The fifth input parameter is optional and describes a set
%# of options useful to adapt the computation to what is needed. 
%# @end itemize
%#
%# @seealso{odepkg}
%# @end deftypefn

function varargout = velocity_verlet (f, t, x, dt, options)

  %# set initial conditions
  dim = length (x) / 2;
  q   = x(1:dim);
  v   = x(dim+1:end);

  a = feval (f, t, x, options.vfunarguments{:})(dim+1:end);
  q_next = q + dt * v + .5 * dt^2 * a;
  a_next = .5 * (a + feval (f, t + dt, [q_next; v],
                            options.vfunarguments{:})(dim+1:end));
  v_next = v +  dt * a_next;

  varargout{1} = t + dt; 
  varargout{2} = [q_next; v_next];

  %# error estimate : apply the same scheme with 2 substeps
  if (nargout >= 3)
    dt_new = dt/2;
    
    ## first substep
    q_next = q + dt_new * v + .5 * dt_new^2 * a;
    a_next = .5 * (a + feval (f, t + dt_new, [q_next; v],
                              options.vfunarguments{:})(dim+1:end));
    v_next = v + dt_new * a_next;
    t = t + dt_new;
    q = q_next;
    v = v_next;

    ## second substep
    a = feval (f, t, [q; v], options.vfunarguments{:})(dim+1:end);
    q_next = q + dt_new * v + .5 * dt_new^2 * a;
    a_next = .5 * ((a + feval (f, t + dt_new, [q_next;v],
                               options.vfunarguments{:}))(dim+1:end));
    v_next = v + dt_new * a_next;

    ## higher accuracy estimates of solution and derivatives
    varargout{3} = [q_next; v_next]; 
    varargout{4} = [];

  endif
  
endfunction

%! Armonic-oscillator
%!function [ydot] = oscillator (vt, vy)
%!  ydot = [vy(length(vy)/2+1:end); -vy(1:length(vy)/2)];
%!endfunction
%!
%!shared opts
%!  opts = odeset;
%!  opts.vfunarguments = {};
%!
%! %# Turn off output of warning messages for all tests, turn them on
%! %# again if the last test is called
%!test
%!  warning ('off', 'OdePkg:InvalidArgument');
%!  [t,y] = velocity_verlet (@oscillator, 0, [1;0], 0.05, opts);
%!test
%!  [t,y,x] = velocity_verlet (@oscillator, 0, [1;1], 0.1, opts);
%!
%!  warning ('on', 'OdePkg:InvalidArgument');

%# Local Variables: ***
%# mode: octave ***
%# End: ***
